"""

Some docstring here!

"""

from riccati.evolve import solve, osc_evolve, nonosc_evolve
from riccati.step import nonosc_step, osc_step
from riccati.stepsize import choose_nonosc_stepsize, choose_osc_stepsize
from riccati.solversetup import solversetup

from . import _version
__version__ = _version.get_versions()['version']
